"""Xns11 has moved to a new repo: https://github.com/DHI/mikeio1d/"""


class Xns11:
    """Xns11 has moved to a new repo: https://github.com/DHI/mikeio1d/"""

    def __init__(self, file_path=None):
        raise NotImplementedError(
            "Xns11 has moved to a new repo: https://github.com/DHI/mikeio1d/"
        )
