"""Res1d has moved to a new repo: https://github.com/DHI/mikeio1d/"""


class Res1d:
    """Res1d has moved to a new repo: https://github.com/DHI/mikeio1d/"""

    def __init__(self, file_path=None, put_chainage_in_col_name=True):
        raise NotImplementedError(
            "Res1d has moved to a new repo: https://github.com/DHI/mikeio1d/"
        )
